/* TERM server: access a local asy device from a tcp connection.
 * Copyright 1991 Phil Karn, KA9Q
 * Adapted from KA9Q's NOS 930622 by N5KNX 11/95.
 */
#include "global.h"
#if defined(TERMSERVER)
#include "netuser.h"
#include "mbuf.h"
#include "socket.h"
#include "proc.h"
#include "remote.h"
#include "smtp.h"
#include "iface.h"
#include "tcp.h"
#include "nr4.h"
#include "commands.h"
#include "hardware.h"
#include "mailbox.h"
#include "asy.h"
#include "i8250.h"
#include "devparam.h"

static void termserv __ARGS((int s,void *unused,void *p));
static void termrx __ARGS((int s,void *p1,void *p2));
static int prompting_read __ARGS((char *prompt, int s, char *buf, int buflen));

static char *Termpass = NULLCHAR;
struct portlist {
    char *portname;
    struct portlist *next;
};
#define NULLPL (struct portlist *)NULL
static struct portlist *Termports = NULLPL;

/* Start up TCP term server */
int
term1(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    int16 port;

    if(argc < 2)
        port = IPPORT_TERM;
    else
        port = atoi(argv[1]);
    return start_tcp(port,"Term Server",termserv,512);
}

/* Stop term server */
int
term0(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    int16 port;

    if(argc < 2)
        port = IPPORT_TERM;
    else
        port = atoi(argv[1]);
    return stop_tcp(port);
}

static void
termserv(s,unused,p)
int s;
void *unused;
void *p;
{
    char buf[64];
    struct iface *ifp;
    struct route *rp;
    struct sockaddr_in fsocket;
    struct proc *rxproc = NULLPROC;
    struct asy *ap;
    struct mbuf *bp;
    struct portlist *pl;
    int (*rawsave) __ARGS((struct iface *,struct mbuf *));
    int i, saved=0;
    
    sockmode(s,SOCK_BINARY);
    sockowner(s,Curproc);
    close_s(Curproc->output);
    close_s(Curproc->input);
    Curproc->output = Curproc->input = s;
    setflush(s, '\n');
    log(s,"open term");

    /* Prompt for and check remote password */
    if (Termpass) {
        if (prompting_read("Password: ", s, buf, sizeof(buf)) < 0)
            goto quit;
        rip(buf);
        if(strcmp(buf,Termpass) != 0){
            tputs("Login incorrect\n");
            goto quit;
        }
    }

    /* Prompt for desired interface. Verify that it exists, that
     * we're not using it for our TCP connection, that it's an
     * asynch port, and that there isn't already another tip, term
     * or dialer session active on it.
     */
    for(;;){
        if (prompting_read("Interface: ",s,buf,sizeof(buf)) < 0)
            goto quit;
        rip(buf);
        if((ifp = if_lookup(buf)) == NULLIF){
            tprintf(Badinterface,buf);
            continue;
        }
        for (pl=Termports; pl; pl=pl->next)
            if (!strcmp(pl->portname, buf)) break;
        if (!pl) {
            tprintf(Badinterface,buf);
            continue;
        }
        if(getpeername(s,(char *)&fsocket,&i) != -1
         && !ismyaddr(fsocket.sin_addr.s_addr)
         && (rp = rt_lookup(fsocket.sin_addr.s_addr)) != NULLROUTE
         && rp->iface == ifp){
            tprintf("You're using interface %s!\n",ifp->name);
            continue;
        }

        ap=&Asy[ifp->dev];
        if( ifp->dev >= ASY_MAX || ap->iface != ifp ){
            tprintf("Interface %s not asy port\n",buf);
            continue;
        }
        if(ifp->raw == bitbucket){
            tprintf("%s already in use\n",buf);
            continue;
        }

        /* Save output handler and temporarily redirect output to null */
        rawsave = ifp->raw;
        ifp->raw = bitbucket;
        saved = 1;

        /* Suspend the packet input driver. Note that the transmit driver
         * is left running since we use it to send buffers to the line.
         */
        suspend(ifp->rxproc);
#ifdef POLLEDKISS
        suspend(ap->poller);
#endif
        break;
    }

    if (prompting_read("Wink DTR? ",s,buf,sizeof(buf)) < 0)
        goto quit;
    if(buf[0] == 'y' || buf[0] == 'Y'){
        asy_ioctl(ifp,PARAM_DTR,1,0);    /* drop DTR */
        pause(1000L);
        asy_ioctl(ifp,PARAM_DTR,1,1);    /* raise DTR */
    }

/* XXX    sockmode(s,SOCK_BINARY);    /* Switch to raw mode */
    if(prompting_read("Turn off local echo? ",s,buf,sizeof(buf)) < 0)
        goto quit;
    if(buf[0] == 'y' || buf[0] == 'Y')
        tprintf("%c%c%c%c%c%c",IAC,WILL,TN_ECHO,IAC,WILL,TN_SUPPRESS_GA);

    /* Now fork into receive and transmit processes */
    rxproc = newproc("term rx",270,termrx,s,ifp,NULL,0);

    /* We continue to handle the TCP->asy direction */
    while((i = recvchar(s)) != EOF){
        while(i == IAC) {
            i = recvchar(s);  /* DO/DONT/WILL/WONT */
            if (i == IAC) break;  /* escaped IAC */
            i = recvchar(s);  /* opt */
            i = recvchar(s);
        }
        if(i == '\n')
            i = '\r';               /* NL => CR */
        bp = pushdown(NULLBUF,1);
        bp->data[0] = (i & 0x7f);
        asy_send(ifp->dev,bp);
        ifp->lastsent = secclock();
        if (i == '\r') pwait(NULL);  /* give output buffers a chance to empty */
    }

quit:
    if (saved) {
        killproc(rxproc);
        ifp->raw = rawsave;
        resume(ifp->rxproc);
#ifdef POLLEDKISS
        resume(ap->poller);
#endif
    }
    log(s,"close term");
    Curproc->input = -1;  /* avoid closing s twice */
    /* close_s(s);  done during process destruction */
}

static void
termrx(int s, void *p1, void *p2)
{
    int c;
    struct iface *ifp = (struct iface *)p1;
    
    while((c = get_asy(ifp->dev)) != EOF)
        usputc(s, c & 0x7f);
    usflush(s);
}

static int
prompting_read(char *prompt, int s, char *buf, int buflen)
{
    int nread;

    if (prompt) {
        usputs(s,prompt);
        usflush(s);
    }
/*    alarm(10000L);*/
    nread = recvline(s,buf,buflen);
/*    alarm(0L);*/

    return nread;
}

int
doterm(int argc,char *argv[],void *p)
{
    int i;
    struct portlist *pl, *pnext;

    if (argc == 1)
usage:
        tputs ("Usage: term ports [portlist] | term password pwstring\n");
    else if (!strcmp(argv[1],"ports")) {
        if (argc == 2) {
            tputs("term ports: ");
            for (pl=Termports; pl; pl=pl->next)
                tprintf("%s ", pl->portname);
            tputc('\n');
        }
        else {
            for (pl=Termports; pl; pl=pnext) {
                free(pl->portname);
                pnext = pl->next;
                free(pl);
            }
            Termports = NULLPL;
            for (i=2; i<argc; i++) {
                pl = mallocw(sizeof (struct portlist));
                pl->portname = strdup(argv[i]);
                pl->next = Termports;
                Termports = pl;
            }
        }
    }
    else if (!strncmp(argv[1], "pass", 4)) {
        if (argc != 3) goto usage;
        free(Termpass);
        Termpass = strdup(argv[2]);
    }
    else goto usage;

    return 0;
}

#endif /* TERMSERVER */
